#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <ctype.h>

#include "parser.h"
#include "extstring.h"

struct SyntaxElement* parse(FILE* file) {
	struct SyntaxElement* syntaxtree = malloc(sizeof(struct SyntaxElement));
	struct SyntaxElement* syntaxtreestart = syntaxtree;
	syntaxtree->type = TOPTREE;
	syntaxtree->next = NULL;
	syntaxtree->content = NULL;
	syntaxtree->top = NULL;
	char symbol;
	bool intree = false;
	bool incomment = false;
	bool instring = false;
	bool intoken = false;
	struct Extstring* token = NULL;
	struct Extstring* string = NULL;
	while (1) {
		symbol = fgetc(file);
		if (feof(file))
			break;
		if (incomment) {
			if (symbol == '\n')
				incomment = false;
			continue;
		}
		if (instring) {
			if (symbol == '"') {
				if (string == NULL) {
					string = malloc(sizeof(struct Extstring));
				}
				instring = false;
				syntaxtree->type = STRING;
				char* sstring = es_tostring(string)->string;
				syntaxtree->content = sstring;
				es_free(string);
				string = NULL;
				continue;
			} else {
				if (string == NULL) {
					string = malloc(sizeof(struct Extstring));
					string->symbol = symbol;
					continue;
				}
				es_addsymbol(string, symbol);
				continue;
			}
		}
		if (intoken) {
			if (isspace(symbol) || symbol == '(' || symbol == ')') {
				intoken = false;
				syntaxtree->type = TOKEN;
				syntaxtree->content = es_tostring(token)->string;
				es_free(token);
				token = NULL;
				continue;
			} else {
				es_addsymbol(token, symbol);
				continue;
			}
		}
		if (isspace(symbol)) 
			continue;
		if (intree) {
			switch (symbol) {
				case '(':
					syntaxtree->content = malloc(sizeof(struct SyntaxElement));
					((struct SyntaxElement*)syntaxtree->content)->top = syntaxtree;
					syntaxtree = syntaxtree->content;
					syntaxtree->content = NULL;
					syntaxtree->next = NULL;
					syntaxtree->type = TREE;
					break;
				case ')':
					if (syntaxtree->top->type == TOPTREE) {
						intree = false;
						syntaxtree = syntaxtree->top;
						syntaxtree->next = malloc(sizeof(struct SyntaxElement));
						syntaxtree = syntaxtree->next;
						syntaxtree->type = TOPTREE;
						syntaxtree->content = NULL;
						syntaxtree->next = NULL;
						syntaxtree->top = NULL;
					} else {
						syntaxtree = syntaxtree->top;
						syntaxtree->next = malloc(sizeof(struct SyntaxElement));
						syntaxtree->next->top = syntaxtree->top;
						syntaxtree = syntaxtree->next;
						syntaxtree->type = TREE;
						syntaxtree->content = NULL;
						syntaxtree->next = NULL;
					}
					break;
				case ';':
					incomment = true;
					continue;
				case '"':
					instring = true;
					continue;
				default:
					token = malloc(sizeof(struct Extstring));
					token->symbol = symbol;
					intoken = true;
					break;
			}
			continue;
		} else {
			switch (symbol) {
				case ')':
					fprintf(stderr, "Unbalanced brackets.\n");
					return NULL;
				case ';':
					incomment = true;
					continue;
				case '(':
					intree = true;
					syntaxtree->content = malloc(sizeof(struct SyntaxElement));
					((struct SyntaxElement*)syntaxtree->content)->top = syntaxtree;
					syntaxtree = syntaxtree->content;
					syntaxtree->content = NULL;
					syntaxtree->next = NULL;
					syntaxtree->type = TREE;
					continue;
				default:
					fprintf(stderr, "Expression outside of brackets.\n");
					return NULL;
			}
		}
	}
	if (intree) {
		fprintf(stderr, "Unexpected EOF.\n");
		return NULL;
	}
	syntaxtree = syntaxtreestart;
	while (1) {
		if (syntaxtree->next->next == NULL) {
			free(syntaxtree->next);
			syntaxtree->next = NULL;
			break;
		}
		syntaxtree = syntaxtree->next;
	}
	return syntaxtreestart;
}
